/* GNOME DB library
 * Copyright (C) 1999-2002 The GNOME Foundation.
 *
 * AUTHORS:
 *      Rodrigo Moya <rodrigo@gnome-db.org>
 *		Daniel Espinosa <esodan@yahoo.com.mx>
 *
 * This Library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this Library; see the file COPYING.LIB.  If not,
 * write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "gnome-db-form-editor.h"
#include "gnome-db-field-editor.h"

#define PARENT_TYPE GTK_TYPE_VBOX

struct _GnomeDbFormEditorPrivate {
	GdaDataModel *model;
	gint current_row;
	GnomeDbFormEditorType type;
	gboolean columned;
	gboolean show_navigator_buttons;

	/* widgets */
	GtkWidget *form_container;

	GList *navigator_widgets; /* navigator widget */
	GtkWidget *navigator_label;
	GtkWidget *navigator_first;
	GtkWidget *navigator_back;
	GtkWidget *navigator_fwd;
	GtkWidget *navigator_last;
};

enum {
	MODEL_CHANGED,
	ROW_CHANGED,
	LAST_SIGNAL
};

static gint form_signals[LAST_SIGNAL];
static GObjectClass *parent_class = NULL;



static void gnome_db_form_editor_class_init   (GnomeDbFormEditorClass *klass);
static void gnome_db_form_editor_init         (GnomeDbFormEditor *form, GnomeDbFormEditorClass *klass);
static void gnome_db_form_editor_set_property (GObject *object,
					guint paramid,
					const GValue *value,
					GParamSpec *pspec);
static void gnome_db_form_editor_get_property (GObject *object,
					guint param_id,
					GValue *value,
					GParamSpec *pspec);
static void gnome_db_form_editor_finalize     (GObject *object);
static void set_row(GnomeDbFormEditor *form, gint row);
static GtkWidget* create_navigator(GnomeDbFormEditor *model);
static GtkWidget* create_list(GnomeDbFormEditor *model);


/*
 * Callbacks
 */

static void
show_navigator_row (GnomeDbFormEditor *form)
{
	gint total_rows;
	gchar *tmp;

	g_return_if_fail (GNOME_DB_IS_FORM_EDITOR (form));

	if (!GDA_IS_DATA_MODEL (form->priv->model)) {
		gtk_label_set_text (GTK_LABEL (form->priv->navigator_label), "No records");
		return;
	}

	if (total_rows <= 0) {
		gtk_widget_set_sensitive (form->priv->navigator_first, FALSE);
		gtk_widget_set_sensitive (form->priv->navigator_back, FALSE);
		gtk_widget_set_sensitive (form->priv->navigator_fwd, FALSE);
		gtk_widget_set_sensitive (form->priv->navigator_last, FALSE);
		gtk_label_set_text (GTK_LABEL (form->priv->navigator_label),"No records");
	}
	else {
		if (form->priv->current_row == 0) {
			gtk_widget_set_sensitive (form->priv->navigator_first, FALSE);
			gtk_widget_set_sensitive (form->priv->navigator_back, FALSE);
		}
		else {
			gtk_widget_set_sensitive (form->priv->navigator_first, TRUE);
			gtk_widget_set_sensitive (form->priv->navigator_back, TRUE);
		}

		if (form->priv->current_row == total_rows - 1) {
			gtk_widget_set_sensitive (form->priv->navigator_fwd, FALSE);
			gtk_widget_set_sensitive (form->priv->navigator_last, FALSE);
		}
		else {
			gtk_widget_set_sensitive (form->priv->navigator_fwd, TRUE);
			gtk_widget_set_sensitive (form->priv->navigator_last, TRUE);
		}

		tmp = g_strdup_printf ("Record %d of %d",
				       form->priv->current_row + 1,
				       total_rows);
		gtk_label_set_text (GTK_LABEL (form->priv->navigator_label), tmp);
		g_free (tmp);
	}
}


static void
navigator_first_cb (GtkButton *button, gpointer user_data)
{
	GnomeDbFormEditor *form = (GnomeDbFormEditor *) user_data;

	g_return_if_fail (GNOME_DB_IS_FORM_EDITOR (form));

	form->priv->current_row = 0;
	show_navigator_row (form);
	g_signal_emit (G_OBJECT (form), form_signals[ROW_CHANGED], 0);
}

static void
navigator_back_cb (GtkButton *button, gpointer user_data)
{
	GnomeDbFormEditor *form = (GnomeDbFormEditor *) user_data;

	g_return_if_fail (GNOME_DB_IS_FORM_EDITOR (form));

	if (form->priv->current_row > 0) {
		form->priv->current_row--;
		show_navigator_row (form);
	}
	g_signal_emit (G_OBJECT (form), form_signals[ROW_CHANGED], 0);
}

static void
navigator_forward_cb (GtkButton *button, gpointer user_data)
{
	gint max_rows;
	GnomeDbFormEditor *form = (GnomeDbFormEditor *) user_data;

	g_return_if_fail (GNOME_DB_IS_FORM_EDITOR (form));

	max_rows = gda_data_model_get_n_rows (form->priv->model) - 1;

	if (form->priv->current_row < max_rows) {
		form->priv->current_row++;
		show_navigator_row (form);
	}
	g_signal_emit (G_OBJECT (form), form_signals[ROW_CHANGED], 0);
}

static void
navigator_last_cb (GtkButton *button, gpointer user_data)
{
	GnomeDbFormEditor *form = (GnomeDbFormEditor *) user_data;

	g_return_if_fail (GNOME_DB_IS_FORM_EDITOR (form));

	form->priv->current_row = gda_data_model_get_n_rows (form->priv->model) - 1;
	show_navigator_row (form);
	g_signal_emit (G_OBJECT (form), form_signals[ROW_CHANGED], 0);
}


/*
 * Private functions
 */


static GtkWidget* 
create_navigator(GnomeDbFormEditor *form)
{
	GtkWidget *box;
	GtkWidget *hbox;
	GtkWidget *vbox;
	GtkWidget *widget;
	GtkWidget *nav_box;
	GtkWidget *navigator;
	GtkWidget *sep;
	GdaValue *value;
	gint col, count;
	
	g_return_val_if_fail (GNOME_DB_IS_FORM_EDITOR (form), NULL);
	g_return_val_if_fail (GDA_IS_DATA_MODEL (form->priv->model), NULL);
	
	/*Create GnomeDbFieldEditor widgets*/
	hbox = gtk_hbox_new(FALSE,0);
	
	vbox = gtk_vbox_new(FALSE, 0);
	gtk_box_pack_start_defaults(GTK_BOX(hbox), vbox);
	
	count = 0;
	for(col = 0 ; col < gda_data_model_get_n_columns(form->priv->model) ; col++)
	{		
		count ++;
		if(count>4)
		{
			vbox = gtk_vbox_new(FALSE, 0);
			gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);
			gtk_widget_show(vbox);
			count = 0;
		}
		
		widget = gnome_db_field_editor_new_with_model (form->priv->model, col, form->priv->current_row);
		
		g_message("Agregando/Mostrando FieldEditor = %s -- FieldEditorCol = %i -- Col = %i", 
						gnome_db_field_editor_get_column_name(
									GNOME_DB_FIELD_EDITOR(widget)), 
						gnome_db_field_editor_get_column(
									GNOME_DB_FIELD_EDITOR(widget)),
						col);
		g_message("Agregando/Mostrando Valor = %s -- RowActual = %i -- Col = %i", 
						gda_value_stringify(gda_data_model_get_value_at(
												form->priv->model,
												col,
												form->priv->current_row)),		
						form->priv->current_row,
						col);
		gtk_box_pack_start(GTK_BOX(vbox), widget, FALSE, FALSE, 10);
		g_object_set_data(G_OBJECT(form), 
							gnome_db_field_editor_get_column_name(
								GNOME_DB_FIELD_EDITOR(widget)),
							widget);
	}
	
	/*Crate Navigator Buttons Bar*/
	navigator = gtk_hbox_new(TRUE, 20);
	if(form->priv->show_navigator_buttons) {
		form->priv->navigator_label = gnome_db_new_label_widget ("No records");
		gtk_misc_set_alignment (GTK_MISC (form->priv->navigator_label), 0.5, 0.5);
		gtk_box_pack_start_defaults(GTK_BOX(navigator),form->priv->navigator_label);
		/* create navigation buttons */
		
		nav_box = gnome_db_new_hbutton_box_widget ();
		gtk_button_box_set_layout (GTK_BUTTON_BOX (nav_box), GTK_BUTTONBOX_DEFAULT_STYLE);
		gtk_box_pack_start_defaults(GTK_BOX(navigator),nav_box);
		
		form->priv->navigator_first = gnome_db_new_button_widget_from_stock (GTK_STOCK_GOTO_FIRST);
		g_signal_connect (G_OBJECT (form->priv->navigator_first), "clicked",
				  G_CALLBACK (navigator_first_cb), form);
		gtk_box_pack_start (GTK_BOX (nav_box), form->priv->navigator_first, FALSE, FALSE, 0);
		
		form->priv->navigator_back = gnome_db_new_button_widget_from_stock (GTK_STOCK_GO_BACK);
		g_signal_connect (G_OBJECT (form->priv->navigator_back), "clicked",
				  G_CALLBACK (navigator_back_cb), form);
		gtk_box_pack_start (GTK_BOX (nav_box), form->priv->navigator_back, FALSE, FALSE, 0);
		
		form->priv->navigator_fwd = gnome_db_new_button_widget_from_stock (GTK_STOCK_GO_FORWARD);
		g_signal_connect (G_OBJECT (form->priv->navigator_fwd), "clicked",
				  G_CALLBACK (navigator_forward_cb), form);
		gtk_box_pack_start (GTK_BOX (nav_box), form->priv->navigator_fwd, FALSE, FALSE, 0);
		
		form->priv->navigator_last = gnome_db_new_button_widget_from_stock (GTK_STOCK_GOTO_LAST);
		g_signal_connect (G_OBJECT (form->priv->navigator_last), "clicked",
				  G_CALLBACK (navigator_last_cb), form);
		gtk_box_pack_start (GTK_BOX (nav_box), form->priv->navigator_last, FALSE, FALSE, 0);
	}
	
	box = gtk_vbox_new(FALSE, 0);
	
	gtk_box_pack_start(GTK_BOX(box), hbox, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(box), navigator, FALSE, FALSE, 0);
	
	gtk_widget_show_all(GTK_WIDGET(box));
	
	show_navigator_row (form);
	
	return box;
}


static GtkWidget *
create_list (GnomeDbFormEditor *form)
{
	g_return_val_if_fail (GNOME_DB_IS_FORM_EDITOR (form), NULL);
	g_return_val_if_fail (GDA_IS_DATA_MODEL (form->priv->model), NULL);

	return gnome_db_new_grid_widget (form->priv->model);
}

static void
setup_widgets_for_model (GnomeDbFormEditor *form)
{
	g_return_if_fail (GNOME_DB_IS_FORM_EDITOR (form));
	/* destroy previous widgets */
	if (GTK_IS_WIDGET(form->priv->form_container))
		gtk_widget_destroy (form->priv->form_container);
	
	/* Delete G_OBJECT's objects saved*/

	/* create the new form depending on the type */
	switch (form->priv->type) {
	case GNOME_DB_FORM_EDITOR_TYPE_NAVIGATOR :
		form->priv->form_container = create_navigator (form);
		break;
	case GNOME_DB_FORM_EDITOR_TYPE_LIST :
		form->priv->form_container = create_list (form);
		break;
	default :
		form->priv->form_container = NULL;
		break;
	}

	if (!form->priv->form_container)
	{
		return;
	}

	gtk_widget_show (form->priv->form_container);
	gtk_box_pack_start (GTK_BOX (form), form->priv->form_container, TRUE, TRUE, 0);
	gtk_widget_show(GTK_WIDGET(form));
}



/*
 * GnomeDbFormEditor class implementation
 */

static void
gnome_db_form_editor_class_init (GnomeDbFormEditorClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	parent_class = g_type_class_peek_parent (klass);

	object_class->set_property = gnome_db_form_editor_set_property;
	object_class->get_property = gnome_db_form_editor_get_property;
	object_class->finalize = gnome_db_form_editor_finalize;

	/* add class signals */
	form_signals[MODEL_CHANGED] =
		g_signal_new ("model_changed",
			      G_TYPE_FROM_CLASS (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (GnomeDbFormEditorClass, model_changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE, 0);
	form_signals[ROW_CHANGED] =
		g_signal_new ("row_changed",
			      G_TYPE_FROM_CLASS (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (GnomeDbFormEditorClass, row_changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID,
			      G_TYPE_NONE, 0);
}

static void
gnome_db_form_editor_init (GnomeDbFormEditor *form, GnomeDbFormEditorClass *klass)
{
	g_return_if_fail (GNOME_DB_IS_FORM_EDITOR (form));

	/* allocate private structure */
	form->priv = g_new0 (GnomeDbFormEditorPrivate, 1);
	form->priv->model = NULL;
	form->priv->type = GNOME_DB_FORM_EDITOR_TYPE_NAVIGATOR;
	form->priv->form_container = NULL;
	form->priv->navigator_label = NULL;
	form->priv->show_navigator_buttons = TRUE;
	form->priv->columned = TRUE;
	form->priv->current_row = 0;
	form->priv->navigator_label = NULL;
	form->priv->navigator_first = NULL;
	form->priv->navigator_back = NULL;
	form->priv->navigator_fwd = NULL;
	form->priv->navigator_last = NULL;
}

static void
gnome_db_form_editor_set_property (GObject *object,
			    guint param_id,
			    const GValue *value,
			    GParamSpec *pspec)
{
	GnomeDbFormEditor *form = (GnomeDbFormEditor *) object;

	g_return_if_fail (GNOME_DB_IS_FORM_EDITOR (form));

	switch (param_id) {
	default :
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
		break;
	}
}

static void
gnome_db_form_editor_get_property (GObject *object,
			    guint param_id,
			    GValue *value,
			    GParamSpec *pspec)
{
	GnomeDbFormEditor *form = (GnomeDbFormEditor *) object;

	g_return_if_fail (GNOME_DB_IS_FORM_EDITOR (form));

	switch (param_id) {
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
		break;
	}
}

static void
gnome_db_form_editor_finalize (GObject *object)
{
	GnomeDbFormEditor *form = (GnomeDbFormEditor *) object;

	g_return_if_fail (GNOME_DB_IS_FORM_EDITOR (form));

	/* free memory */
	if (GDA_IS_DATA_MODEL (form->priv->model)) {
		g_object_unref (G_OBJECT (form->priv->model));
		form->priv->model = NULL;
	}

	if (form->priv->navigator_widgets) {
		g_list_free (form->priv->navigator_widgets);
		form->priv->navigator_widgets = NULL;
	}

	g_free (form->priv);
	form->priv = NULL;

	/* chain to parent class */
	parent_class->finalize (object);
}

GType
gnome_db_form_editor_get_type (void)
{
	static GType type = 0;

	if (!type) {
		static const GTypeInfo info = {
			sizeof (GnomeDbFormEditorClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) gnome_db_form_editor_class_init,
			NULL,
			NULL,
			sizeof (GnomeDbFormEditor),
			0,
			(GInstanceInitFunc) gnome_db_form_editor_init
		};
		type = g_type_register_static (PARENT_TYPE, "GnomeDbFormEditor", &info, 0);
	}
	return type;
}

/**
 * gnome_db_form_new
 *
 * Create a new #GnomeDbFormEditor widget, which is a high-level widget which
 * lets you display recordsets in many different ways, such as a list,
 * record-by-record, etc.
 *
 * Returns: the newly created widget.
 */
GtkWidget *
gnome_db_form_editor_new (void)
{
	GnomeDbFormEditor *form;

	form = g_object_new (GNOME_DB_TYPE_FORM_EDITOR, NULL);
	return GTK_WIDGET (form);
}

GtkWidget*
gnome_db_form_editor_new_with_model(GdaDataModel *model, gint row, gboolean columned) 
{
	GnomeDbFormEditor *form;
	
	form = GNOME_DB_FORM_EDITOR(gnome_db_form_editor_new());
		
	form->priv->model = model;
	g_object_ref(model);
	form->priv->current_row = row;
	form->priv->columned = columned;
	setup_widgets_for_model(form);
	return form;
}

GtkWidget*
gnome_db_form_editor_get_field (GnomeDbFormEditor *form, const gchar *name) 
{
	return GTK_WIDGET(g_object_get_data(G_OBJECT(form), name));
}

void
gnome_db_form_editor_update_data(GnomeDbFormEditor *form)
{
	GnomeDbFieldEditor *field;
	GdaRow *row;
	GdaValue *val;
	gint i;
	gboolean changed;
		
	row = gda_data_model_get_row(form->priv->model, form->priv->current_row);
		
	for(i = 0; i < gda_data_model_get_n_columns(form->priv->model); i++)
	{
		val = gda_row_get_value(row, i);
		field = (GnomeDbFieldEditor*) 
					g_object_get_data(G_OBJECT(form),
							gda_data_model_get_column_title(form->priv->model, i));
		if(GNOME_DB_IS_FIELD_EDITOR(field))
		{
			GdaValue *v = gnome_db_field_editor_get_widget_value(field);
			if(gda_value_compare(v, val) != 0)	
			{
				/* If diferent from the stored in the FieldEditor*/
				changed = TRUE;
				gda_value_set_from_value(val, v);
			}
			g_message("Original Value: \'%s\' Value to update: \'%s\' Actualizar=%i\n", 
									gda_value_stringify(val), gda_value_stringify(v), changed);
		}
		else
			g_message("En UpdateDataForm: No es un FIELD\n");
	}
	
	if(changed)
	{
		gda_data_model_begin_update(form->priv->model);
		gda_data_model_update_row(form->priv->model, row);
		gda_data_model_end_update(form->priv->model);
	}
}

void
gnome_db_form_editor_delete_current_row (GnomeDbFormEditor *form)
{
	GdaRow *row = gda_data_model_get_row(form->priv->model, form->priv->current_row);
	
	gda_data_model_remove_row(form->priv->model, row);
}

void
gnome_db_form_editor_set_row(GnomeDbFormEditor *form, gint row)
{
	form->priv->current_row = row;
	set_row(form, row);
}

static void
set_row(GnomeDbFormEditor *form, gint row)
{
	gint col;
	GnomeDbFieldEditor *field_data;
	
	form->priv->current_row = row;
	
	for(col=0; col < gda_data_model_get_n_columns(form->priv->model); col ++)
	{
		field_data = GNOME_DB_FIELD_EDITOR(g_object_get_data(G_OBJECT(form), 
									gda_data_model_get_column_title(
														form->priv->model, col)));
		if(GNOME_DB_IS_FIELD_EDITOR(field_data))
			gnome_db_field_editor_set_row(field_data, row);
		
	}
}

void
gnome_db_form_editor_free(GnomeDbFormEditor *form)
{
	g_free(form);
}

GtkWidget*
gnome_db_form_editor_find_field (GnomeDbFormEditor *form, const gchar *name)
{
	return GTK_WIDGET(g_object_get_data(G_OBJECT(form), name));	
}

/**
 * gnome_db_form_editor_set_model
 * @form: a #GnomeDbForm widget.
 * @model: the data model to associate the widget with.
 */
void
gnome_db_form_set_model (GnomeDbFormEditor *form, GdaDataModel *model)
{
	g_return_if_fail (GNOME_DB_IS_FORM_EDITOR (form));
	g_return_if_fail (GDA_IS_DATA_MODEL (model));

	if (GDA_IS_DATA_MODEL (form->priv->model))
		g_object_unref (G_OBJECT (form->priv->model));

	g_object_ref (G_OBJECT (model));
	form->priv->model = model;
	setup_widgets_for_model (form);
	g_signal_emit (G_OBJECT (form), form_signals[MODEL_CHANGED], 0);
}
